(* MultiInst - Multiinstance helper component
 * Copyright (C) 2001 Tomas Mandys-MandySoft
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *)

{ URL: http://www.2p.cz }

{
  TMultiInstance comoponent controls behavior of multiple intances of the same program.
  Default allows only one instance of program. Next instance terminates immediatly.

  Starting program sends global broadcast message where wParam is unique id.
  All listening programs are handling messages in WndProc and when own AppId and wParam
  equal - it means just is started new instance - returns to starting program -1.
  If SendMessage returns -1 program terminates otherwise is filled InstId property.
  Instance having lowest InstId is master of instances and controls responses to new
  instances. In basic TMultiInstance is allowed only one instance of program . More
  complex functionality (limited number of instances, ...) is possible add when overriding
  DoOnStartInstance, DoOnCloseInstance methods.

  MultiMessage name registered in initialization section and AppId MUST BE UNIQUE
  among applications running on computer. To do it when distributing applications
  change name of multi message to unique name, f.ex. an GUID (Ctrl+Shift+G)

  (Do)OnForceClose is called in new instance when old/master instance disallowed new one. Abort discards termination.
  (Do)OnStartInstance is called in old/master instance when new one starting
  (Do)OnCloseInstance is called in old/master instance when priviously allowed multi instance is closing (possible master change)
   OnGetAppId is called when is necessary to change dynamically AppId in run-time

  CheckMethod determines what method use when checking if running anothed instance
    mimMessage .. SendMessage and waits for response .. disadvantage: possible lock in messages loop
    mimFileMapping .. CreateFileMapping used, no events served on master side
}

unit MultiInst;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs;

const
  DevelIdentifier = 'TMultiInstance message-MandySoft';

type
  TMultiInstanceEvent = procedure(Sender: TObject; var fInstId: Integer) of object;
  TMultiInstanceGetAppId = procedure(Sender: TObject; var fAppId: Integer) of object;
  TMultiCheckMethod = (mimMessage, mimFileMapping);

  TMultiInstance = class(TComponent)
  private
    FWindowHandle: HWND;
    fAppId: Integer;
    fInstId: Integer;
    fMapping: THandle;
    fCheckMethod: TMultiCheckMethod;
    fOnForceClose: TNotifyEvent;
    fOnStartInstance: TMultiInstanceEvent;
    fOnCloseInstance: TMultiInstanceEvent;
    fOnGetAppId: TMultiInstanceGetAppId;
    procedure WndProc(var Msg: TMessage);
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    property InstId: Integer read fInstId;
    function GetNewInstId: Integer;
  protected
    procedure DoOnForceClose; virtual;
    procedure DoOnStartInstance(var aInstId: Integer); virtual;
    procedure DoOnCloseInstance(aInstId: Integer); virtual;
    procedure Loaded; override;
    function GetMasterHandle: HWND;
    function GetAppId: Integer; virtual;
  published
    property AppId: Integer read fAppId write fAppId default 1;
    property CheckMethod: TMultiCheckMethod read fCheckMethod write fCheckMethod default mimMessage;
    property OnForceClose: TNotifyEvent read fOnForceClose write fOnForceClose;
    property OnStartInstance: TMultiInstanceEvent read fOnStartInstance write fOnStartInstance;
    property OnCloseInstance: TMultiInstanceEvent read fOnCloseInstance write fOnCloseInstance;
    property OnGetAppId: TMultiInstanceGetAppId read fOnGetAppId write fOnGetAppId; 
  end;

var
  hMultiMessage: UINT;

procedure Register;

implementation

constructor TMultiInstance.Create(AOwner:TComponent);
begin
  inherited;
  if not (csDesigning in ComponentState) then
    FWindowHandle := AllocateHWnd(WndProc);
  fAppId:= 1;
  fCheckMethod:= mimMessage;
end;

destructor TMultiInstance.Destroy;
begin
  if not (csDesigning in ComponentState) then
  begin
    if fInstId <> 0 then  { destroying correctly created instance }
    begin
      PostMessage(HWND_BROADCAST, hMultiMessage, fInstId, GetAppId); { notification }
    end;
    if fMapping <> 0 then
      CloseHandle(fMapping);
    DeallocateHWnd(FWindowHandle);
  end;
  inherited Destroy;
end;

procedure TMultiInstance.DoOnCloseInstance;
begin
  if Assigned(fOnCloseInstance) then
    fOnCloseInstance(Self, aInstId);
end;

procedure TMultiInstance.DoOnForceClose;
begin
  try
    if Assigned(fOnForceClose) then
      fOnForceClose(Self);
    Application.Terminate;   
  except
    on E: Exception do
      if not (ExceptObject is EAbort) then
        raise;
  end;
end;

procedure TMultiInstance.DoOnStartInstance;
begin
  aInstId:= -1;   { to be closed new created instance }
  if Assigned(fOnStartInstance) then
    fOnStartInstance(Self, aInstId);
end;

type
  PEnumPar = ^TEnumPar;
  TEnumPar = record
    Handle: HWND;
    AppId: Integer;
    InstId: Integer;
    Flag: Boolean;
  end;

function EnumWinProc(aHandle: HWND; Info: PEnumPar): BOOL; stdcall;
var
  Res: Integer;
begin
  Result:= True;
  Res:= SendMessage(aHandle, hMultiMessage, -1, Info^.AppId);
  if Res > 0 then
    if Info^.Flag then
      begin
        if Res > Info^.InstId then
        begin
          Info^.InstId:= Res;
          Info^.Handle:= aHandle;
        end
      end
    else
      if Res < Info^.InstId then
      begin
        Info^.InstId:= Res;
        Info^.Handle:= aHandle;
        Result:= Res <> 1;  { 1 is 100% master }
      end;
end;

function TMultiInstance.GetAppId: Integer;
begin
  Result:= fAppId;
  if Assigned(fOnGetAppId) then
    fOnGetAppId(Self, Result);
end;

function TMultiInstance.GetMasterHandle: HWND;
var
  Info: TEnumPar;
begin
  Info.Handle:= 0;
  Info.AppId:= GetAppId;
  Info.InstId:= $FFFF;  { longint x integer }
  Info.Flag:= False;
  EnumWindows(@EnumWinProc, Longint(@Info));
  Result:= Info.Handle;
end;

function TMultiInstance.GetNewInstId: Integer;
var
  Info: TEnumPar;
begin
  Info.Handle:= 0;
  Info.AppId:= GetAppId;
  Info.InstId:= 1;
  Info.Flag:= True;
  EnumWindows(@EnumWinProc, Longint(@Info));
  Result:= Info.InstId;
end;

procedure TMultiInstance.Loaded;
var
  MHandle: HWND;
begin
  if csDesigning in ComponentState then
    Exit;
  if fCheckMethod = mimMessage then
    begin
      MHandle:= GetMasterHandle();
      if MHandle = 0 then   { no master, it is first instance }
        fInstId:= 1
      else
        begin
          fInstId:= SendMessage(MHandle, hMultiMessage, 0, GetAppId); { ask to master }
          if fInstId = -1 then  { instance already exists and not allowed next instance by first/master instance }
            DoOnForceClose
        end;
    end
  else if fCheckMethod = mimFileMapping then
    begin
      fMapping:= CreateFileMapping($FFFFFFFF, nil, PAGE_READONLY, 0, 32, PChar(DevelIdentifier+'_'+IntToStr(GetAppId)));
      if fMapping <> 0 then
        begin
          if GetLastError() = ERROR_ALREADY_EXISTS then
            DoOnForceClose; 
        end
      else
       //  "Error creating mapping", "DevelIdentifier+'_'+IntToStr(GetAppId)"
       ;
    end;
end;

procedure TMultiInstance.WndProc(var Msg: TMessage);
begin
  if (Msg.Msg = hMultiMessage) and (Msg.lParam = GetAppId) and (Msg.wParam <> fInstId) then
    begin
      if Msg.wParam = -1 then
        Msg.Result:= fInstId
      else if Msg.wParam <> 0 then
        DoOnCloseInstance(Msg.Result)
      else
        DoOnStartInstance(Msg.Result);
    end
  else
    Msg.Result := DefWindowProc(FWindowHandle, Msg.Msg, Msg.wParam, Msg.lParam);
end;

procedure Register;
begin
  RegisterComponents('System', [TMultiInstance]);
end;

begin
  hMultiMessage:= RegisterWindowMessage(DevelIdentifier);
end.
